function [samples, fpred] = NonParSF(target, data, pred, cost, rstar, Mn0, Meta, Msd, burnin, thin, numbofits)

lambda = - 1 / log(rstar);

firm = data(:, size(data, 2));
data = data(:, 1:(size(data, 2)-1));

[n p] = size(data);

numbfirms = max(firm);

for i = 1:numbfirms
    if ( sum(firm==i) == 0 )
        'Error in firm indices'
        exit(1)
    end
end

data = [ones(n, 1) data];

p = p + 1;

s = (1:numbfirms)';
numb = ones(numbfirms, 1);
u = - log(rand(max(s), 1)) / lambda;
M = 1;

lambda0 = 10^(-6);

xtx = data' * data;

holdsigmasq = zeros(numbofits, 1);
holdbeta = zeros(numbofits, p);
holdlambda = zeros(numbofits, 1);
holdu = zeros(numbofits, numbfirms);
holdM = zeros(numbofits, 1);
fpred = zeros(1, length(pred));

Maccept = 0;
Mcount = 0;

beta = inv(data' * data) * data' * target;
sigmasq = var(target -  data * beta);

totalits = burnin + thin * numbofits;

for it = 1:totalits

    % Updating beta
    varstar = inv(xtx + sigmasq * lambda0 * eye(p));
    if ( cost == 1 )
        xty = data' * (target - u(s(firm)));
    else
        xty = data' * (target + u(s(firm)));
    end
    meanstar = varstar * xty;

    beta = meanstar + sqrt(sigmasq) * chol(varstar)' * randn(p, 1);

    if ( cost == 1 )
        alphastar = - log(rand) / (lambda * length(numb));
        minu = min(u);
        u = u - minu + alphastar;
        beta(1) = beta(1) + minu - alphastar;
    else
        oldbeta1 = beta(1);
        beta(1) = beta(1) - min(u) - log(rand) / (lambda * length(numb));
        u = u + beta(1) - oldbeta1;
    end
        
    % Updating s
    for i = 1:numbfirms
        numb(s(i)) = numb(s(i)) - 1;
        if ( numb(s(i)) == 0 )
            numb(s(i), :) = [];
            u(s(i), :) = [];
            s = s - (s > s(i));
        end

        prob = zeros(1, length(numb) + 1);

        err = target(firm==i) - data(firm==i, :) * beta;
        nstar = sum(firm==i);
        if ( cost == 1 )
            for j = 1:length(numb)
                prob(j) = numb(j)*exp(- 0.5 * sum((err - u(j)).^2) / sigmasq);
            end
            prob(length(numb)+1) = M*lambda*sqrt(sigmasq/nstar)*exp(- 0.5 * (sum(err.^2) - (sum(err) - sigmasq * lambda)^2/nstar) / sigmasq);
        else
            for j = 1:length(numb)
                prob(j) = numb(j)*exp(- 0.5 * sum((err + u(j)).^2) / sigmasq);
            end
            prob(length(numb)+1) = M*lambda*sqrt(sigmasq/nstar)*exp(- 0.5 * (sum(err.^2) - (sum(err) + sigmasq * lambda)^2/nstar) / sigmasq);
        end
        fprob = cumsum(prob);
        fprob = fprob ./ fprob(end);

        ustar = rand;
        s(i) = 1;
        while ( fprob(s(i)) < ustar )
            s(i) = s(i) + 1;
        end

        if ( s(i) == length(u) + 1 )
            numb = [numb; 1];
            if ( cost == 1 )
                mustar = (sum(err) - sigmasq * lambda) / nstar;
            else
                mustar = - (sum(err) + sigmasq * lambda) / nstar;
            end
            varstar = sigmasq / nstar;
            newu = truncated_normal(mustar, varstar, 0, inf);
            u = [u; newu];
        else
            numb(s(i)) = numb(s(i)) + 1;
        end
    end

    % Updating u
    sum1 = zeros(length(numb), 1);
    sum2 = zeros(length(numb), 1);
    for i = 1:n
        sum1(s(firm(i))) = sum1(s(firm(i))) + target(i) - data(i, :) * beta;
        sum2(s(firm(i))) = sum2(s(firm(i))) + 1;
    end

    for i = 1:length(u)
        if ( cost == 1 )
            mustar = - (sum1(i) + sigmasq * lambda) / sum2(i);
        else
            mustar = (sum1(i) - sigmasq * lambda) / sum2(i);
        end
        varstar = sigmasq / sum2(i);
        u(i) = truncated_normal(mustar, varstar, 0, inf);
    end

    % Updating sigmasq
    if ( cost == 1 )
        sigmasq = 1 / gamrnd(0.001 + 0.5 * n, 1 / (0.001 + 0.5 * sum((target - u(s(firm)) - data * beta).^2)));
    else
        sigmasq = 1 / gamrnd(0.001 + 0.5 * n, 1 / (0.001 + 0.5 * sum((target + u(s(firm)) - data * beta).^2)));
    end
    
    % Updating lambda
    lambda = gamrnd(length(u) + 1, 1 / (sum(u) - log(rstar)));

    % Updating M
    newM = M * exp(Msd * randn);

    logaccept = length(numb) * log(newM) + gammaln(newM) - gammaln(newM + numbfirms);
    logaccept = logaccept - (length(numb) * log(M) + gammaln(M) - gammaln(M + numbfirms));
    logaccept = logaccept + Meta * log(newM) - 2 * Meta * log(newM + Mn0);
    logaccept = logaccept - (Meta * log(M) - 2 * Meta * log(M + Mn0));

    accept = 1;
    if ( logaccept < 0 )
        accept = exp(logaccept);
    end

    Maccept = Maccept + accept;
    Mcount = Mcount + 1;

    if ( rand < accept )
        M = newM;
    end

    if ( mod(it, 100) == 0 )
        it
        Maccept / Mcount
    end

    if ( (it > burnin) && (mod(it - burnin, thin) == 0))
        holdsigmasq((it - burnin) / thin) = sigmasq;
        holdbeta((it - burnin) / thin, :) = beta;
        holdlambda((it - burnin) / thin) = lambda;
        holdu((it - burnin) / thin, :) = u(s);
        holdM((it - burnin) / thin) = M;
        for i = 1:length(pred);
            fpred(i) = fpred(i) + sum(numb(u<pred(i))) / (M + numbfirms) + M / (M + numbfirms) * (1 - exp(- lambda * pred(i)));
        end
    end
end

fpred = fpred / numbofits;

samples = struct('u',holdu,'beta',holdbeta,'lambda',holdlambda,'sigmasq',holdsigmasq,'M',holdM);